const fs = require("fs");
const tmp = require("tmp");
const stream = require("stream");
const unzipper = require("unzipper");
const Sax = require("sax");
const Util = require("util");
const ssf = require("ssf");
const saxStream = require("sax-stream");
const request = require("request");
const validUrl = require("valid-url");

function localFileSize(filePath) {
  return new Promise(function (resolve, reject) {
    if (!fs.existsSync(filePath)) {
      resolve(-400);
    } else {
      fs.stat(filePath, function (error, stats) {
        if (error) {
          resolve(-300);
        } else if (!stats) {
          resolve(-200);
        } else {
          resolve(stats.size);
        }
      });
    }
  });
}

function remoteFileSize(options) {
  // From https://github.com/evanlucas/remote-file-size
  if ("string" === typeof options) {
    options = {
      uri: options
    };
  }
  options = options || {};

  options.method = "HEAD";
  options.followAllRedirects = true;
  options.followOriginalHttpMethod = true;

  return new Promise(function (resolve, reject) {
    request(options, function (err, res, body) {
      if (err) return cb(err);
      const code = res.statusCode;
      if (code >= 400) {
        // return cb(new Error('Received invalid status code: ' + code))
        resolve(-400);
      }

      var len = res.headers["content-length"];
      if (!len) {
        // return cb(new Error('Unable to determine file size'))
        resolve(-300);
      }
      len = +len;
      if (len !== len) {
        // return cb(new Error('Invalid Content-Length received'))
        resolve(-200);
      }

      resolve(+res.headers["content-length"]);
    });
  });
}

function getFileSize(file) {
  if (validUrl.isUri(file)) {
    return remoteFileSize(file);
  }
  return localFileSize(file);
}

function getReadStream(file) {
  if (validUrl.isUri(file)) {
    return request(file);
  }
  return fs.createReadStream(file);
}

function memoryUsage() {
  const used = process.memoryUsage().heapUsed / 1024 / 1024;
  console.log(
    `The script uses approximately ${Math.round(used * 100) / 100} MB`
  );
}

var Transform = stream.Transform || require("readable-stream").Transform;

/*******************************************************************************
 Extract Contents(Sheetes) from Excel
 ******************************************************************************/
function ReadWorkBookStream(
  cachedSharedStrings,
  cachedFormats,
  sharedOptions,
  options
) {
  var self = this;
  if (!(this instanceof ReadWorkBookStream)) {
    return new ReadWorkBookStream(
      cachedSharedStrings,
      cachedFormats,
      sharedOptions,
      options
    );
  }

  // init Transform
  if (!options) options = {}; // ensure object
  options.objectMode = true; // forcing object mode
  Transform.call(this, options);
  this.parsedSharedStrings = false;
  this.parsedStyles = false;
  this.sharedStrings = cachedSharedStrings;
  this.formatNumbers = [];
  this.formats = cachedFormats;
  this.styleStream;
  if (!sharedOptions) sharedOptions = {};
  this.destroyStatus = sharedOptions.destroyStatus;
  this.pauseStatus = sharedOptions.pauseStatus;
  this.timeout = sharedOptions.timeOut;
  this.sheetIndexes = self._indexes(sharedOptions.sheetIndexes);
  this.waitingWorkSheets = sharedOptions.waitingWorkSheets;
}

Util.inherits(ReadWorkBookStream, Transform);

ReadWorkBookStream.prototype._indexes = function (indexes) {
  var self = this;
  if (typeof indexes === "string") {
    return [parseInt(indexes)];
  } else if (typeof indexes === "number") {
    return [indexes];
  } else if (Array.isArray(indexes)) {
    return indexes;
  }
  return [1];
};

ReadWorkBookStream.prototype._transform = function (entry, encoding, done) {
  var self = this;

  if (self.abortBook) {
    entry.autodrain();
  }

  switch (entry.path) {
    case "xl/workbook.xml":
      entry.autodrain();
      break;
    case "xl/_rels/workbook.xml.rels":
      entry.autodrain();
      break;
    case "_rels/.rels":
      entry.autodrain();
      break;
    case "xl/sharedStrings.xml":
      self._parseSharedStrings(entry);
      break;
    case "xl/styles.xml":
      self._parseStyles(entry);
      break;
    default:
      if ((match = entry.path.match(/xl\/(worksheets\/sheet(\d+)\.xml)/))) {
        var sheetPath = match[1];
        var sheetNo = match[2];
        if (self.sheetIndexes.includes(parseInt(sheetNo))) {
          if (!self.parsedSharedStrings && !self.parsedStyles) {
            const { name } = tmp.fileSync({});
            const stream = fs.createWriteStream(name);
            self.waitingWorkSheets.push({ sheetNo: sheetNo, path: name });
            entry.pipe(stream);
          } else {
            self.push(entry);
          }
        }
      } else if (
        (match = entry.path.match(
          /xl\/worksheets\/_rels\/sheet(\d+)\.xml.rels/
        ))
      ) {
        entry.autodrain();
      } else {
        entry.autodrain();
      }
      break;
  }
  done();
};

ReadWorkBookStream.prototype._parseSharedStrings = function (
  sharedStringsStream
) {
  var self = this;
  sharedStringsStream
    .pipe(
      saxStream({
        strict: true,
        tag: "si"
      })
    )
    .on("data", function (x) {
      if (x.children.t) {
        self.sharedStrings.push(x.children.t.value);
      } else {
        let str = "";
        for (let i = 0; i < x.children.r.length; i++) {
          const ch = x.children.r[i].children;
          str += ch.t.value;
        }
        self.sharedStrings.push(str);
      }
    })
    .on("end", function (x) {
      self.parsedSharedStrings = true;
    })
    .on("error", function (error) {
      self.emit("error", error);
    });
};

ReadWorkBookStream.prototype._parseStyles = function (styleStream) {
  var self = this;
  styleStream
    .pipe(
      saxStream({
        strict: true,
        tag: ["cellXfs", "numFmts"]
      })
    )
    .on("data", function (x) {
      if (x.tag === "numFmts" && x.record.children) {
        const children = x.record.children.numFmt.length
          ? x.record.children.numFmt
          : [x.record.children.numFmt];
        for (let i = 0; i < children.length; i++) {
          self.formatNumbers[Number(children[i].attribs.numFmtId)] =
            children[i].attribs.formatCode;
        }
      } else if (x.tag === "cellXfs" && x.record.children) {
        for (let i = 0; i < x.record.children.xf.length; i++) {
          const ch = x.record.children.xf[i];
          self.formats[i] = Number(ch.attribs.numFmtId);
        }
      }
    })
    .on("end", function (x) {
      for (let i = 0; i < self.formats.length; i++) {
        const format = self.formatNumbers[self.formats[i]];
        if (format) {
          self.formats[i] = format;
        }
      }
      self.parsedStyles = true;
    })
    .on("error", function (error) {
      self.emit("error", error);
    });
};

ReadWorkBookStream.prototype._flush = function (done) {
  var self = this;
  if (self.waitingWorkSheets.length) {
    for (let index = 0; index < self.waitingWorkSheets.length; index++) {
      var sheetInfo = self.waitingWorkSheets[index];
      var sheetNo = sheetInfo.sheetNo;
      var workSheetStream = fs.createReadStream(sheetInfo.path);
      self.push(workSheetStream);
    }
  }
  done();
};

ReadWorkBookStream.prototype._final = function (done) {
  done();
};

ReadWorkBookStream.prototype._cleanupTmp = function (done) {
  var self = this;
  self.waitingWorkSheets.forEach((element) => {
    if (fs.existsSync(element.path)) {
      fs.unlinkSync(element.path);
    }
  });
};

/*******************************************************************************
 ReadSheetStream
 ******************************************************************************/
function ReadWorkSheetStream(
  cachedSharedStrings,
  cachedFormats,
  sharedOptions,
  options
) {
  var self = this;
  if (!(this instanceof ReadWorkSheetStream)) {
    return new ReadWorkSheetStream(
      cachedSharedStrings,
      sharedOptions,
      cachedFormats,
      options
    );
  }

  // init Transform
  if (!options) options = {}; // ensure object
  options.objectMode = true; // forcing object mode
  Transform.call(this, options);
  this.sharedStrings = cachedSharedStrings;
  this.formats = cachedFormats;
  if (!sharedOptions) sharedOptions = {};
  this.destroyStatus = sharedOptions.destroyStatus;
  this.pauseStatus = sharedOptions.pauseStatus;
  this.timeout = sharedOptions.timeOut;
  this.workSheetStream;
  this.batchSize = sharedOptions.batchSize;
  this.hasHeader = sharedOptions.hasHeader;
  this.index = 0;
  this.records = 0;
  this.sax = saxStream({ strict: true, tag: "row" });
}

Util.inherits(ReadWorkSheetStream, Transform);

ReadWorkSheetStream.prototype._transform = function (
  workSheetStream,
  encoding,
  done
) {
  var self = this;
  self.workSheetStream = workSheetStream;
  self.workSheetStream
    .pipe(self.sax)
    .on("data", function (row) {
      if (!self.destroyStatus || !self.destroyStatus.onDestroy) {
        self._pushData(row);
      } else {
        self.__endAll();
      }
    })
    .on("end", function () {
      done();
    })
    .on("error", function (error) {
      self.emit("error", error);
    });

  if (self.destroyStatus && self.destroyStatus.onDestroy) {
    self.__endAll();
  }
};

ReadWorkSheetStream.prototype._pushData = function (row) {
  var self = this;
  self.records = self.index === 0 && self.hasHeader ? 0 : self.records + 1;
  if (self.batchSize && self.batchSize > 1 && self.records === self.batchSize) {
    self.records = 0;
    self.sax.pause();
    self.workSheetStream.pause();
    setTimeout(() => {
      // self.resume();
      self.workSheetStream.resume();
      self.sax.resume();
    }, self.timeout);
  }
  self.push(row);
  self.index++;
};

ReadWorkSheetStream.prototype.__endAll = function () {
  self.sax.end();
  self.workSheetStream.end();
  self.end();
};

ReadWorkSheetStream.prototype._flushVars = function () {
  var self = this;
  self.index = 0;
  self.records = 0;
};
/*******************************************************************************
 ReadRowStream
 ******************************************************************************/
function ReadWorkRowStream(
  cachedSharedStrings,
  cachedFormats,
  sharedOptions,
  options
) {
  var self = this;
  if (!(this instanceof ReadWorkRowStream)) {
    return new ReadWorkRowStream(
      cachedSharedStrings,
      cachedFormats,
      sharedOptions,
      options
    );
  }
  // init Transform
  if (!options) options = {}; // ensure object
  options.objectMode = true; // forcing object mode
  Transform.call(this, options);
  this.sharedStrings = cachedSharedStrings;
  this.formats = cachedFormats;
  if (!sharedOptions) sharedOptions = {};
  this.destroyStatus = sharedOptions.destroyStatus;
  this.pauseStatus = sharedOptions.pauseStatus;
  this.timeout = sharedOptions.timeOut;
  this.sliceSingle = false;
  this.sliceRange = [];
  this.slicingEnable = this._canSlice(sharedOptions.slice);
}

Util.inherits(ReadWorkRowStream, Transform);

ReadWorkRowStream.prototype._lettersToNumber = function (letters) {
  return letters.split("").reduce((r, a) => r * 26 + parseInt(a, 36) - 9, 0);
};

ReadWorkRowStream.prototype._canSlice = function (slice) {
  var self = this;
  if (typeof slice === "number" && slice > 0) {
    self.sliceRange = [0, slice];
  } else if (typeof slice === "string" && parseInt(slice) > 0) {
    self.sliceRange = [0, parseInt(slice)];
  } else if (
    Array.isArray(slice) &&
    slice.length === 2 &&
    slice[0] >= 0 &&
    slice[1] > 0
  ) {
    self.sliceRange =
      slice[1] <= slice[0] ? [slice[0], slice[0] + slice[1]] : slice;
  }

  if (
    self.sliceRange &&
    self.sliceRange.length &&
    self.sliceRange[1] - self.sliceRange[0] === 1
  ) {
    self.sliceSingle = true;
  }

  return self.sliceRange && self.sliceRange.length > 0;
};

ReadWorkRowStream.prototype._slice = function (array) {
  var self = this;
  if (self.sliceSingle) {
    return array[self.sliceRange[0]];
  } else if (self.slicingEnable) {
    return array.slice(self.sliceRange[0], self.sliceRange[1]);
  }
  return array;
};

ReadWorkRowStream.prototype._transform = function (chunk, encoding, done) {
  var self = this;
  let arr = [];
  const children = chunk.children
    ? chunk.children.c.length
      ? chunk.children.c
      : [chunk.children.c]
    : [];
  for (let i = 0; i < children.length; i++) {
    const ch = children[i];
    if (ch.children) {
      let value = ch.children.v.value;
      if (ch.attribs.t === "s") {
        value = self.sharedStrings[value];
      }
      value = isNaN(value) ? value : Number(value);
      let column = ch.attribs.r.replace(/[0-9]/g, "");
      const index = self._lettersToNumber(column) - 1;
      // format
      const formatId = ch.attribs.s ? Number(ch.attribs.s) : 0;
      if (formatId) {
        const format = self.formats[formatId];
        value =
          format && format.toString().includes("#")
            ? value
            : ssf.format(self.formats[formatId], value);
        value = isNaN(value) ? value : Number(value);
      }
      arr[index] = value;
    }
  }

  if (self.destroyStatus && self.destroyStatus.onDestroy) {
    done();
    self.end();
  } else {
    done(
      undefined,
      !arr.length ? null : self._slice(Array.from(arr, (x) => (x ? x : "")))
    );
  }
};

/*******************************************************************************
 ReadLineStream
 ******************************************************************************/
function ReadWorkLineStream(sharedOptions, options) {
  var self = this;
  if (!(this instanceof ReadWorkLineStream)) {
    return new ReadWorkLineStream(sharedOptions, options);
  }
  // init Transform
  if (!options) options = {}; // ensure object
  options.objectMode = true; // forcing object mode
  Transform.call(this, options);
  this.index = 0;
  this.records = [];
  if (!sharedOptions) sharedOptions = {};
  this.transformCallbackFn = sharedOptions.transformCallbackFn;
  this.endCallbackFn = sharedOptions.endCallbackFn;
  this.headerCallBackFn = sharedOptions.headerCallBackFn;
  this.parsingCallbackFn = sharedOptions.parsingCallbackFn;
  this.batchSize = sharedOptions.batchSize;
  this.timeout = sharedOptions.timeOut;
  this.destroyStatus = sharedOptions.destroyStatus;
  this.pauseStatus = sharedOptions.pauseStatus;
  this.waitingWorkSheets = sharedOptions.waitingWorkSheets;
  this.hasHeader = sharedOptions.hasHeader;
  this.memoryUsage = sharedOptions.memoryUsage;
}

Util.inherits(ReadWorkLineStream, Transform);

ReadWorkLineStream.prototype._transform = function (chunk, encoding, done) {
  var self = this;
  var action = "";
  if (
    self.index === 0 &&
    typeof self.headerCallBackFn === "function" &&
    self.hasHeader
  ) {
    action = "GET_HEADER";
  } else if (!self.batchSize || self.batchSize < 2) {
    action = "NO_BATCH";
  } else {
    action = "BATCH";
  }

  switch (action) {
    case "GET_HEADER":
      self.headerCallBackFn(chunk);
      break;

    case "NO_BATCH":
      self._parsing(self._customTranformation(chunk));
      break;

    case "BATCH":
      self.records.push(self._customTranformation(chunk));
      if (self.records.length === self.batchSize) {
        if (self.memoryUsage) memoryUsage();
        self._parsing(self.records, true);
        self.records = [];
      }
      break;

    default:
      break;
  }

  self.index++;

  if (self.destroyStatus && self.destroyStatus.onDestroy) {
    done();
    self.end();
  } else {
    done();
  }
};

ReadWorkLineStream.prototype._flush = function (done) {
  var self = this;
  if (self.records.length > 0) {
    self._parsing(self.records, true);
  }

  if (typeof self.endCallbackFn === "function") {
    self.endCallbackFn();
  }

  self._cleanupTmp();
  self.index = 0;
  self.records = [];

  self.emit("endWorkline");
  done();
};

ReadWorkLineStream.prototype._customTranformation = function (element) {
  var self = this;
  return typeof self.transformCallbackFn === "function"
    ? self.transformCallbackFn(element)
    : element;
};

ReadWorkLineStream.prototype._parsing = function (elements, batch) {
  var self = this;
  return typeof self.parsingCallbackFn === "function"
    ? self.parsingCallbackFn(elements, batch)
    : elements;
};

ReadWorkLineStream.prototype._cleanupTmp = function (done) {
  var self = this;
  self.waitingWorkSheets.forEach((element) => {
    if (fs.existsSync(element.path)) {
      fs.unlinkSync(element.path);
    }
  });
};

/*******************************************************************************
 Main class
 ******************************************************************************/

function ExcelStreamParser() {}

/*
options = {
  file: file,
  parsingCallbackFn: parsingCallbackFn,
  batchSize: batchSize,
  transformCallbackFn: transformCallbackFn,
  endCallbackFn: endCallbackFn,
  headerCallBackFn: headerCallBackFn,
  timeOut: timeOut,
  destroyStatus: destroyStatus,
  sheetIndex: sheetIndex,
  skipLine: skipLine,
  parseInBatch: parseInBatch,
  slice: slice,
  errorCallbackFn: errorCallbackFn,
  hasHeader: hasHeader,
  memoryUsage: memoryUsage
  }
*/

ExcelStreamParser.prototype.parse = function parse(options) {
  if (!options) options = {};

  var sharedStrings = [];
  var formats = [];
  var waitingWorkSheets = [];

  var pauseStatus = {
    onPause: false
  };

  const sharedOptions = {
    transformCallbackFn: options.transformCallbackFn,
    endCallbackFn: options.endCallbackFn,
    headerCallBackFn: options.headerCallBackFn,
    parsingCallbackFn: options.parsingCallbackFn,
    excelFileErrorCallBackFn: options.excelFileErrorCallBackFn,
    batchSize: options.batchSize,
    timeOut:
      options.timeOut && typeof options.timeOut === "number"
        ? options.timeOut
        : 500,
    destroyStatus: options.destroyStatus,
    pauseStatus: pauseStatus,
    waitingWorkSheets: waitingWorkSheets,
    hasHeader: options.hasHeader,
    slice: options.slice,
    memoryUsage: options.memoryUsage
  };

  var readWorkBookStream = new ReadWorkBookStream(
    sharedStrings,
    formats,
    sharedOptions
  );
  var readWorkSheetStream = new ReadWorkSheetStream(
    sharedStrings,
    formats,
    sharedOptions
  );
  var readWorkRowStream = new ReadWorkRowStream(
    sharedStrings,
    formats,
    sharedOptions
  );
  var readWorkLineStream = new ReadWorkLineStream(sharedOptions);
  var zipParser = unzipper.Parse();

  const stream = getReadStream(options.file);

  stream
    .pipe(zipParser)
    .pipe(readWorkBookStream)
    .pipe(readWorkSheetStream)
    .pipe(readWorkRowStream)
    .pipe(readWorkLineStream);

  zipParser.on("error", function (error) {
    if (typeof options.excelFileErrorCallBackFn === "function") {
      options.excelFileErrorCallBackFn(error);
    }
  });

  stream
    .on("end", function () {
      sharedStrings = [];
      formats = [];
    })
    .on("error", function (error) {
      console.log("Errororor");
      if (typeof options.errorCallbackFn === "function") {
        options.errorCallbackFn(error);
      }
    });

  readWorkLineStream.on("endWorkline", function () {
    readWorkSheetStream._flushVars();
  });
};

module.exports = ExcelStreamParser;
