const fs = require("fs");
const { Transform } = require("stream");
const excel = require("@adncorp/excel-stream");
const request = require("request");
const validUrl = require("valid-url");
var csv = require("csv-stream");

function getReadStream(file) {
  if (validUrl.isUri(file)) {
    return request(file);
  }
  return fs.createReadStream(file);
}

function memoryUsage() {
  const used = process.memoryUsage().heapUsed / 1024 / 1024;
  console.log(`The script uses approximately ${Math.round(used * 100) / 100} MB`);
}

function ExcelParser() {}

ExcelParser.prototype.parse = function parse(options) {
  if(!options) options = {};

  const bSize = options.batchSize && typeof options.batchSize === "number" ? options.batchSize : 10000;
  // stream of input file
  const stream = getReadStream(options.file);
  const tOut = options.timeOut && typeof options.timeOut === "number" ? options.timeOut : 100;
  const sheetId = options.sheetIndex ? options.sheetIndex : 0;
  const skip = options.skipLine ? options.skipLine : 0;
  // convert excel to object stream
  stream
    .pipe(excel({ sheetIndex: sheetId, fieldSep: ";", delimiter: ";" }))
    .pipe(
      getTransformObject(
        options.parsingCallbackFn,
        bSize,
        options.transformCallbackFn,
        options.endCallbackFn,
        options.headerCallBackFn,
        tOut,
        options.destroyStatus,
        skip,
        options.parseInBatch,
        options.slice
      )
    );
};

// return transform stream which will do processing
function getTransformObject(
  parsingCallbackFn,
  batchSize,
  transformCallbackFn,
  endCallbackFn,
  headerCallBackFn,
  timeOut,
  destroy,
  skip,
  parseInBatch,
  sliceLength
) {
  const jsonToDb = new Transform({
    writableObjectMode: true,
    readableObjectMode: true,
    transform(chunk, encoding, callback) {
      if (this.index === 0 && typeof headerCallBackFn === "function") {
        headerCallBackFn(Object.keys(chunk));
      }
      var trans = this;
      this.index++;
      if (destroy && destroy.value) {
        callback();
        this.end();
      } else if (this.index <= skip) {
        callback();
      } else {
        this.records.push(
          sliceLength
            ? Object.values(chunk).slice(0, sliceLength)
            : Object.values(chunk)
        );
        // batch processing of records
        if (this.records.length === batchSize) {
          memoryUsage();
          this.pause();
          applyCustomTransformation(
            this.records,
            transformCallbackFn,
            timeOut
          ).then(data => {
            if (!parseInBatch) {
              data.map(record => parsingCallbackFn(record));
            } else {
              parsingCallbackFn(data);
            }
            data = null;
            this.records = [];
            setTimeout(() => {
              trans.resume();
              callback();
            }, 1000);
          });
        } else {
          callback();
        }
      } 
    },
    flush(done) {
      if (this.records.length > 0) {
        applyCustomTransformation(
          this.records,
          transformCallbackFn,
          timeOut
        ).then(data => {
          if (!parseInBatch) {
            data.map(record => parsingCallbackFn(record));
          } else {
            parsingCallbackFn(data);
          }
          data = null;
          this.records = [];
          done();
          if (typeof endCallbackFn === "function") {
            endCallbackFn();
          }
        });
      } else {
        done();
        if (typeof endCallbackFn === "function") {
          endCallbackFn();
        }
      }
    }
  });
  jsonToDb.records = [];
  jsonToDb.index = 0;

  return jsonToDb;
}

// async function to process data
function applyCustomTransformation(elements, transformCallbackFn, timeOut) {
  return new Promise((resolve, reject) => {
    setTimeout(() => {
      resolve(
        elements.map(
          element =>
            typeof transformCallbackFn === "function"
              ? transformCallbackFn(element)
              : element
        )
      );
    }, timeOut);
  });
}

module.exports = ExcelParser;
